"""Astronomy utility module.

This module provides methods to convert among astronomical quantities
like luminosity and magnitude.

Utility functions
=================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   luminosity_from_absolute_magnitude
   absolute_magnitude_from_luminosity

"""

import numpy as np


def luminosity_from_absolute_magnitude(absolute_magnitude):
    """ Converts absolute magnitudes into luminosities

    Parameters
    ----------
    absolute_magnitude : array_like
                    Input absolute magnitudes
    Returns
    -------
    ndarray, or float if input is scalar
    Luminosity values.
    """

    return np.power(10, -0.4*absolute_magnitude)


def absolute_magnitude_from_luminosity(luminosity):
    """ Converts luminosities into absolute magnitudes

    Parameters
    ----------
    luminosity : array_like
            Input luminosity

    Returns
    -------
    ndarray, or float if input is scalar
    Absolute magnitude values
    """
    return -np.log(luminosity)/(0.4 * np.log(10))



from astropy.cosmology import default_cosmology
from astropy.io import fits
from astropy.table import Table
import networkx
import numpy as np
import os
import pytest
from skypy.pipeline.driver import SkyPyDriver


def test_driver():

    # Evaluate and store the default astropy cosmology.
    config = {'cosmology': {
                'module': 'astropy.cosmology',
                'function': 'default_cosmology.get'}}

    driver = SkyPyDriver()
    driver.execute(config)
    assert driver.cosmology == default_cosmology.get()

    # Generate a simple two column table with a dependency. Also write the
    # table to a fits file and check it's contents.
    config = {'tables': {
                'test_table': {
                  'column1': {
                    'module': 'numpy.random',
                    'function': 'uniform',
                    'kwargs': {
                      'size': 100}},
                  'column2': {
                    'module': 'numpy.random',
                    'function': 'uniform',
                    'requires': {
                      'low': 'test_table.column1'}}}}}

    driver = SkyPyDriver()
    driver.execute(config, file_format='fits')
    assert len(driver.test_table) == 100
    assert np.all(driver.test_table['column1'] < driver.test_table['column2'])
    with fits.open('test_table.fits') as hdu:
        assert np.all(Table(hdu[1].data) == driver.test_table)

    # Check for failure if 'column1' requires itself creating a cyclic
    # dependency graph
    config['tables']['test_table']['column1']['requires'] = \
        {'low': 'test_table.column1'}
    with pytest.raises(networkx.NetworkXUnfeasible):
        driver.execute(config)

    # Check for failure if 'column1' and 'column2' both require each other
    # creating a cyclic dependency graph
    config['tables']['test_table']['column1']['requires'] = \
        {'low': 'test_table.column2'}
    with pytest.raises(networkx.NetworkXUnfeasible):
        driver.execute(config)

    # Check for failure if 'column1' is removed from the config so that the
    # requirements for 'column2' are not satisfied.
    del config['tables']['test_table']['column1']
    with pytest.raises(KeyError):
        driver.execute(config)


def teardown_module(module):

    # Remove fits file generated in test_driver
    os.remove('test_table.fits')



"""Special functions.

This module computes useful special functions.


Utility functions
=================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   upper_incomplete_gamma

"""

import numpy as np

import scipy.special


def upper_incomplete_gamma(a, x):
    """ Non-regularised upper incomplete gamma function. Extension of the
    regularised upper incomplete gamma function implemented in SciPy. In
    this way you can pass a negative value for a.

    Parameters
    ----------
    a : array_like
        Parameter
    x : array_like
        Nonnegative parameter

    Returns
    -------
    Scalar or ndarray
        Value of the non-regularised upper incomplete gamma function.
    """
    if a > 0:
        return scipy.special.gammaincc(a, x) * scipy.special.gamma(a)
    return (scipy.special.gammaincc(a + 1, x)
            - np.power(x, a) * np.exp(-x) / scipy.special.gamma(a + 1)) \
        * scipy.special.gamma(a)



r"""Galaxy spectrum module.


Models
======

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   dirichlet_coefficients

"""

import numpy as np


def dirichlet_coefficients(redshift, alpha0, alpha1, z1=1.):
    r"""Dirichlet-distributed SED coefficients.

    Spectral coefficients to calculate the rest-frame spectral energy
        distribution of a galaxy following the Herbel et al. model in [1]_.

    Parameters
    ----------
    redshift : (nz,) array_like
        The redshift values of the galaxies for which the coefficients want to
        be sampled.
    alpha0, alpha1 : (nc,) array_like
        Factors parameterising the Dirichlet distribution according to Equation
        (3.9) in [1]_.
    z1 : float or scalar, optional
       Reference redshift at which alpha = alpha1. The default value is 1.0.

    Returns
    -------
    coefficients : (nz, nc) ndarray
        The spectral coefficients of the galaxies. The shape is (n, nc) with nz
        the number of redshifts and nc the number of coefficients.

    Notes
    -----
    One example is the rest-frame spectral energy distribution of galaxies
    which can be written as a linear combination of the five kcorrect [2]_
    template spectra :math:`f_i`
    (see [1]_)

    .. math::

        f(\lambda) = \sum_{i=1}^5 c_i f_i(\lambda) \;,

    where the coefficients :math:`c_i` were shown to follow a Dirichlet
    distribution of order 5. The five parameters describing the Dirichlet
    distribution are given by

    .. math::

        \alpha_i(z) = (\alpha_{i,0})^{1-z/z_1} \cdot (\alpha_{i,1})^{z/z_1} \;.

    Here, :math:`\alpha_{i,0}` describes the galaxy population at redshift
    :math:`z=0` and :math:`\alpha_{i,1}` the population at :math:`z=z_1 > 0`.
    These parameters depend on the galaxy type and we chose :math:`z_1=1`.

    Beside this example, this code works for a general number of templates.

    References
    ----------
    .. [1] Herbel J., Kacprzak T., Amara A. et al., 2017, Journal of Cosmology
           and Astroparticle Physics, Issue 08, article id. 035 (2017)
    .. [2] Blanton M. R., Roweis S., 2007, The Astronomical Journal,
           Volume 133, Page 734

    Examples
    --------
    >>> from skypy.galaxy.spectrum import dirichlet_coefficients
    >>> import numpy as np

    Sample the coefficients according to [1]_ for n blue galaxies with
    redshifts between 0 and 1.

    >>> n = 100000
    >>> alpha0 = np.array([2.079, 3.524, 1.917, 1.992, 2.536])
    >>> alpha1 = np.array([2.265, 3.862, 1.921, 1.685, 2.480])
    >>> redshift = np.linspace(0,2, n)
    >>> coefficients = dirichlet_coefficients(redshift, alpha0, alpha1)

    """
    if np.isscalar(alpha0) or np.isscalar(alpha1):
        raise ValueError("alpha0 and alpha1 must be array_like.")
    return_shape = (*np.shape(redshift), *np.shape(alpha0))
    redshift = np.atleast_1d(redshift)[:, np.newaxis]

    alpha = np.power(alpha0, 1. - redshift / z1) * \
        np.power(alpha1, redshift / z1)

    # To sample Dirichlet distributed variables of order k we first sample k
    # Gamma distributed variables y_i with the parameters alpha_i.
    # Normalising all y_i by the sum of all k y_i gives us the k Dirichlet
    # distributed variables.
    y = np.random.gamma(alpha)
    sum_y = y.sum(axis=1)
    coefficients = np.divide(y.T, sum_y.T).T
    return coefficients.reshape(return_shape)



r'''Halo mass module.

This module provides methods to sample the  masses of dark matter halos.

Models
======

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   press_schechter

'''

import numpy as np

from skypy.utils.random import schechter


def press_schechter(n, m_star, size=None, x_min=0.00305,
                    x_max=1100.0, resolution=100):
    """Sampling from Press-Schechter mass function (1974).

    Masses following the Press-Schechter mass function following the
    Press et al. [1]_ formalism.

    Parameters
    ----------
    n : float or int
        The n parameter in the Press-Schechter mass function.
    m_star : float or int
        Factors parameterising the characteristic mass.
    size: int, optional
         Output shape of luminosity samples.
    x_min, x_max : float or int, optional
        Lower and upper bounds in units of M*.
    resolution : int, optional
        Resolution of the inverse transform sampling spline. Default is 100.

    Returns
    -------
    mass : array_like
        Drawn masses from the Press-Schechter mass function.

    Examples
    --------
    >>> import skypy.halo.mass as mass
    >>> n, m_star = 1, 1e9
    >>> sample = mass.press_schechter(n, m_star, size=1000, x_min=1e-10,
    ...                               x_max=1e2, resolution=1000)

    References
    ----------
    .. [1] Press, W. H. and Schechter, P., APJ, (1974).

    """

    alpha = - 0.5 * (n + 9.0) / (n + 3.0)

    x_sample = schechter(alpha, x_min, x_max, resolution=resolution, size=size)

    return m_star * np.power(x_sample, 3.0 / (n + 3.0))



from astropy.units import allclose, isclose
import numpy as np
from scipy import stats
from scipy.stats.tests.common_tests import (
    check_normalization, check_moment, check_mean_expect, check_var_expect,
    check_skew_expect, check_kurt_expect, check_edge_support,
    check_random_state_property, check_pickling)


def test_beta_ellipticity():

    from skypy.galaxy.ellipticity import beta_ellipticity

    # Initialise a randomised ellipticity distribution, an equivalent beta
    # distribution and special cases where the ellipticity distribution is
    # equivalent to a uniform distribution and an arcsine distribution.
    a, b = np.random.lognormal(size=2)
    args = (a / (a + b), a + b)
    beta_dist = stats.beta(a, b)
    ellipticity_dist = beta_ellipticity(*args)
    ellipticity_uniform = beta_ellipticity(0.5, 2.0)
    ellipticity_arcsine = beta_ellipticity(0.5, 1.0)

    # Range of input values spanning the support of the distributions
    x = np.linspace(0, 1, 100)

    # Check basic properties of distribution implementation
    check_normalization(beta_ellipticity, args, 'beta_ellipticity')
    check_edge_support(beta_ellipticity, args)
    check_random_state_property(beta_ellipticity, args)
    check_pickling(beta_ellipticity, args)

    # Check distribution moments
    m, v, s, k = ellipticity_dist.stats(moments='mvsk')
    check_mean_expect(beta_ellipticity, args, m, 'beta_ellipticity')
    check_var_expect(beta_ellipticity, args, m, v, 'beta_ellipticity')
    check_skew_expect(beta_ellipticity, args, m, v, s, 'beta_ellipticity')
    check_kurt_expect(beta_ellipticity, args, m, v, k, 'beta_ellipticity')
    check_moment(beta_ellipticity, args, m, v, 'beta_ellipticity')

    # Compare ellipticity distribution functions (e.g. pdf, cdf...) against
    # functions for an equivalent beta distribution
    assert allclose(ellipticity_dist.pdf(x), beta_dist.pdf(x))
    assert allclose(ellipticity_dist.logpdf(x), beta_dist.logpdf(x))
    assert allclose(ellipticity_dist.cdf(x), beta_dist.cdf(x))
    assert allclose(ellipticity_dist.logcdf(x), beta_dist.logcdf(x))
    assert allclose(ellipticity_dist.ppf(x), beta_dist.ppf(x))
    assert allclose(ellipticity_dist.sf(x), beta_dist.sf(x))
    assert allclose(ellipticity_dist.logsf(x), beta_dist.logsf(x))
    assert allclose(ellipticity_dist.isf(x), beta_dist.isf(x))
    assert isclose(ellipticity_dist.entropy(), beta_dist.entropy())
    assert isclose(ellipticity_dist.median(), beta_dist.median())
    assert isclose(ellipticity_dist.std(), beta_dist.std())
    assert allclose(ellipticity_dist.interval(x), beta_dist.interval(x))

    # Test scalar output
    assert np.isscalar(ellipticity_dist.rvs())

    # Test array output
    assert ellipticity_dist.rvs(size=10).shape == (10,)

    # Test broadcast output
    e_ratio = 0.5 * np.ones((13, 1, 5))
    e_sum = 0.5 * np.ones((7, 5))
    rvs = beta_ellipticity.rvs(e_ratio, e_sum)
    assert rvs.shape == np.broadcast(e_ratio, e_sum).shape

    # Kolmogorov-Smirnov test comparing ellipticity and beta distributions
    D, p = stats.kstest(ellipticity_dist.rvs, beta_dist.cdf, N=1000)
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # Kolmogorov-Smirnov test comparing ellipticity and uniform distributions
    D, p = stats.kstest(ellipticity_uniform.rvs, 'uniform', N=1000)
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # Kolmogorov-Smirnov test comparing ellipticity and arcsine distributions
    D, p = stats.kstest(ellipticity_arcsine.rvs, 'arcsine', N=1000)
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)



import numpy as np
from astropy.cosmology import Planck15
from astropy.units import allclose
from astropy.utils.data import get_pkg_data_filename
from unittest.mock import patch, MagicMock

# load the external camb result to test against
camb_result_filename = get_pkg_data_filename('data/camb_result.txt')
mock_pkz = np.loadtxt(camb_result_filename, delimiter=',')

# create a mock object and specify values for all the attributes needed in
# camb.py
camb_mock = MagicMock()
camb_result = [0, 1, mock_pkz.T]
camb_mock.get_results().get_matter_power_spectrum.return_value = camb_result

# try to import the requirement, if it doesn't exist, use the mock instead
try:
    __import__('camb')
    camb_import_loc = {}
except ImportError:
    camb_import_loc = {'camb': camb_mock}


@patch.dict('sys.modules', camb_import_loc)
def test_camb():
    '''
    Test a default astropy cosmology
    '''
    from skypy.power_spectrum import camb

    redshift = [0.0, 1.0]
    wavenumber = np.logspace(-4.0, np.log10(2.0), 200)
    pkz = camb(wavenumber, redshift, Planck15, 2.e-9, 0.965)
    assert pkz.shape == (len(wavenumber), len(redshift))
    assert allclose(pkz, mock_pkz, rtol=1.e-4)

    # also check redshifts are ordered correctly
    redshift = [1.0, 0.0]
    pkz = camb(wavenumber, redshift, Planck15, 2.e-9, 0.965)
    assert pkz.shape == (len(wavenumber), len(redshift))
    assert allclose(pkz, mock_pkz[:, np.argsort(redshift)], rtol=1.e-4)



import numpy as np
import scipy.stats
import scipy.integrate


import skypy.halo.mass as mass


def test_press_schechter():
    # Test the schechter function, sampling dimensionless x values
    n, m_star = 1, 1e9
    alpha = - 0.5 * (n + 9.0) / (n + 3.0)

    def calc_cdf(m):
        x = np.power(m/m_star, 1 + n/3)
        pdf = np.power(x, alpha) * np.exp(- x)
        cdf = scipy.integrate.cumtrapz(pdf, x, initial=0)
        cdf = cdf / cdf[-1]
        return cdf

    sample = mass.press_schechter(n, m_star, x_min=1e-1, x_max=1e2,
                                  resolution=100, size=1000)

    # Test the distribution of galaxy mass follows the right distribution
    p_value = scipy.stats.kstest(sample, calc_cdf)[1]
    assert p_value > 0.01



# Autogenerated by Astropy-affiliated package skypy's setup.py on 2020-07-14 19:42:05 UTC
from __future__ import unicode_literals
import datetime

version = "0.1.1rc1"
githash = "70a4649272c7fb48ed26f740ab127470bbea9ab5"


major = 0
minor = 1
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 7, 14, 19, 42, 5)
debug = False

astropy_helpers_version = "3.2.1"



import skypy.utils.special as special


def test_upper_incomplete_gamma():
    assert round(special.upper_incomplete_gamma(0.5, 1.5), 6) == 0.147583
    assert round(special.upper_incomplete_gamma(-0.3, 1.5), 6) == 0.080007



"""Galaxy ellipticity module.

This module provides facilities to sample the ellipticities of galaxies.


Models
======

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   beta_ellipticity

"""

from scipy import stats


def _ellipticity_parameterization(beta_method):
    def reparameterize(self, *args):
        *beta_args, e_ratio, e_sum = args
        a = e_sum * e_ratio
        b = e_sum * (1.0 - e_ratio)
        return beta_method(self, *beta_args, a, b)
    return reparameterize


class beta_ellipticity_gen(stats._continuous_distns.beta_gen):
    r'''Galaxy ellipticities sampled from a reparameterized beta distribution.

    The ellipticities follow a beta distribution parameterized by
    :math:`e_{\rm ratio}` and :math:`e_{\rm sum}` as presented in [1]_ Section
    III.A.

    Parameters
    ----------
    e_ratio : array_like
        Mean ellipticity of the distribution, must be between 0 and 1.
    e_sum : array_like
        Parameter controlling the width of the distribution, must be positive.

    Notes
    -----
    The probability distribution function :math:`p(e)` for ellipticity
    :math:`e` is given by a beta distribution:

    .. math::

        p(e) \sim \frac{\Gamma(a+b)}{\Gamma(a) \Gamma(b)} x^{a-1} (1-x)^{b-1}

    for :math:`0 <= e <= 1`, :math:`a = e_{\rm sum} e_{\rm ratio}`,
    :math:`b = e_{\rm sum} (1 - e_{\rm ratio})`, :math:`0 < e_{\rm ratio} < 1`
    and :math:`e_{\rm sum} > 0`, where :math:`\Gamma` is the gamma function.

    References
    ----------
    .. [1] Kacprzak T., Herbel J., Nicola A. et al., arXiv:1906.01018

    Examples
    --------
    >>> from skypy.galaxy.ellipticity import beta_ellipticity

    Sample 10 random variates from the Kacprzak model with
    :math:`e_{\rm ratio} = 0.5` and :math:`e_{\rm sum} = 1.0`:

    >>> ellipticity = beta_ellipticity.rvs(0.5, 1.0, size=10)

    Fix distribution parameters for repeated use:

    >>> ellipticity_distribution = beta_ellipticity(0.5, 1.0)
    >>> ellipticity_distribution.mean()
    0.5
    >>> ellipticity = ellipticity_distribution.rvs(size=10)
    '''

    @_ellipticity_parameterization
    def _rvs(self, *args):
        return super()._rvs(*args)

    @_ellipticity_parameterization
    def _logpdf(self, *args):
        return super()._logpdf(*args)

    @_ellipticity_parameterization
    def _cdf(self, *args):
        return super()._cdf(*args)

    @_ellipticity_parameterization
    def _ppf(self, *args):
        return super()._ppf(*args)

    @_ellipticity_parameterization
    def _stats(self, *args):
        return super()._stats(*args)

    def fit(self, data, *args, **kwargs):
        return super(stats._continuous_distns.beta_gen, self).fit(
            data, *args, **kwargs)


beta_ellipticity = beta_ellipticity_gen(a=0.0, b=1.0, name='beta_ellipticity',
                                        shapes="e_ratio e_sum")



"""Driver module.

This module provides methods to run pipelines of functions with dependencies
and handle their results.
"""

from astropy.table import Table
import networkx
import re


__all__ = [
    'SkyPyDriver',
]


class SkyPyDriver:
    r'''Class for running pipelines.

    This is the main class for running pipelines of functions with dependencies
    and using their results to generate tables.
    '''

    def execute(self, config, file_format=None):
        r'''Run a pipeline.

        This function runs a pipeline of functions to generate the cosmology
        and the columns of a set of tables. It uses a Directed Acyclic Graph to
        determine a non-blocking order of execution that resolves any
        dependencies, see [1]_. Tables can optionally be written to file.

        Parameters
        ----------
        config : dict-like
            Configuration for the pipeline.
        file_format : str
            File format used to write tables. Files are written using the
            Astropy unified file read/write interface; see [2]_ for supported
            file formats. If None (default) tables are not written to file.

        Notes
        -----
        Each step in the pipeline is configured by a dictionary specifying:

        - 'function' : the name of the function
        - 'module' : the name of the the module to import 'function' from
        - 'args' : a list of positional arguments (by value)
        - 'kwargs' : a dictionary of keyword arguments
        - 'requires' : a dictionary of keyword arguments

        Note that 'kwargs' specifices keyword arguments by value, wheras
        'requires' specifices the names of previous steps in the pipeline and
        uses their return values as keyword arguments.

        'config' should contain 'cosmology' and/or 'tables'. 'cosmology' should
        return a dictionary configuring a function that returns an
        astropy.cosmology.Cosmology object. 'tables' should contain a set of
        nested dictionaries, first giving the names of each table, then the
        names of each column within each table. Each column should return a
        dictionary configuring a function that returns an array-like object.

        See [3]_ for examples of pipeline configurations in yaml format.

        References
        ----------
        .. [1] https://networkx.github.io/documentation/stable/
        .. [2] https://docs.astropy.org/en/stable/io/unified.html
        .. [3] https://github.com/skypyproject/skypy/tree/master/examples
        '''

        # Create a Directed Acyclic Graph of all jobs and dependencies
        dag = networkx.DiGraph()
        if 'cosmology' in config:
            dag.add_node('cosmology')
        for table, columns in config.get('tables', {}).items():
            dag.add_node(table)
            for column, settings in columns.items():
                node = '.'.join((table, column))
                dag.add_node(node)
        for table, columns in config.get('tables', {}).items():
            for column, settings in columns.items():
                node = '.'.join((table, column))
                dag.add_edge(table, node)
                requirements = settings.get('requires', {}).values()
                dag.add_edges_from((r, node) for r in requirements)

        # Execute jobs in order that resolves dependencies
        for job in networkx.topological_sort(dag):
            if job == 'cosmology':
                settings = config.get('cosmology')
                self.cosmology = self._call_from_config(settings)
            elif job in config.get('tables', {}):
                setattr(self, job, Table())
            else:
                table, column = job.split('.')
                settings = config['tables'][table][column]
                getattr(self, table)[column] = self._call_from_config(settings)

        # Write tables to file
        if file_format:
            for table in config.get('tables', {}).keys():
                filename = '.'.join((table, file_format))
                getattr(self, table).write(filename)

    def _call_from_config(self, config):

        # Import function
        module_name = config.get('module')
        object_name, function_name = re.search(r'^(\w*?)\.?(\w*)$',
                                               config.get('function')).groups()
        if object_name:
            module = __import__(module_name, fromlist=object_name)
            object = getattr(module, object_name)
            function = getattr(object, function_name)
        else:
            module = __import__(module_name, fromlist=function_name)
            function = getattr(module, function_name)

        # Parse arguments
        args = config.get('args', [])
        kwargs = config.get('kwargs', {})
        req = config.get('requires', {})
        req = {k: self.__getitem__(v) for k, v in req.items()}

        # Call function
        return function(*args, **kwargs, **req)

    def __getitem__(self, label):
        name, key = re.search(r'^(\w*)\.?(\w*)$', label).groups()
        item = getattr(self, name)
        return item[key] if key else item
